(* AUTEURS / AUTHORS :                                     *)
(* Yuxin Deng, Jiatong University, Shanghai                *)
(* Jean-François Monin, Verimag, Université Grenoble Alpes *)
(*                                             2010 - 2019 *)

Require Import Arith.
Require Import List.
Require Import Nat.
Require Import Recdef.

Set Implicit Arguments.

(* ------------------------------------------------------------------------------ *)

Require Import list_facts.
Require Import finsets.
Require Import dec_facts.
Require Import filter_lt.

Require Import notations.

Require Import statement_defs.

(* ------------------------------------------------------------------------------ *)
Require Import rat_facts.
(* ------------------------------------------------------------------------------ *)


Section sec_probas.
(*----------- Input --------------------------*)

Variable Prb : ℕ -> basic_test -> Q.
(* Hypothesis: probabilities are non-negative. *)
Hypothesis not_neg_bt : ∀ i t, 0 <= Prb i t.

(* Initial set *)
Variable I0 : set ℕ.
Hypothesis initially_non_empty : I0 <> ∅.

Variable oracle : ∀ i j : ℕ, (i, j) ∈ I0×I0 -> distinguish Prb i j.

(* Shorthand *)
Definition Pr := Prtest Prb.

(* ------------------------ *)
(* Basic facts about probas *)

Lemma not_neg_t : ∀ i t, 0 <= Pr i t.
Proof.
  induction t; simpl.
  - intuition.
  - apply Qmult_le_0_compat. apply not_neg_bt. assumption.
Qed.

Lemma not_neg_test : ∀ i t, 0 <= Pr i t.
Proof.
  induction t.
  - unfold Pr. intuition.
  - cbn. generalize (Qmult_le_compat_r 0 (Prb i a) _ (not_neg_bt _ _ ) IHt).
    generalize (Qmult_0_l (Pr i t)). intros H1 H2. rewrite H1 in H2. assumption.
Qed.

Lemma Pr_cons : ∀ i bt t, Pr i (bt ::t ) == (Prb i bt) * (Pr i t).
Proof. reflexivity. Qed.

(* ------------------------ *)

Definition filter_pos (t : basic_test) := filter (fun x => is_pos (Prb x t)).

Definition eq_prob (t : test) (x : ℕ²) :=
  let (x1, x2) := x in Qeq_bool (Pr x1 t) (Pr x2 t).

Definition filter_eq (t : test) := filter (eq_prob t).

Definition eq_prob_bt (t : basic_test) (x : ℕ²) :=
  let (x1, x2) := x in Qeq_bool (Prb x1 t) (Prb x2 t).

Definition filter_neq (t : basic_test) := filter (fun a => negb (eq_prob_bt t a)).

(* *)

Lemma rep_bt : ∀ n bt t i,
  Pr i ((repeat bt n) ++ t) == (natpow (Prb i bt) n) * Pr i t.
Proof.
  intros *. induction n as [| n IHn]; simpl.
  - rewrite Qmult_1_l. reflexivity.
  - rewrite IHn. apply Qmult_assoc.
Qed.

(* *)

Record in_iter_data : Type := mk_in_iter_data {
  Y : set ℕ²;
  Y_term : set ℕ²;
  tt : test;
}.

(* Y_pass is constant in the inner loop *)
Section sec_with_Y_pass.
Variable Y_pass : set ℕ².
Hypothesis ndYpass : NoDup Y_pass.

Inductive in_data_invariant (d: in_iter_data) : Prop :=
| in_data_invariant_intro :
    Y d ⊆ Y_pass -> Y_term d ⊆ Y_pass ->
    NoDup (Y d) -> NoDup (Y_term d) -> in_data_invariant d.

Definition next_Y_inner_loop r := filter_eq (tt r) (Y_pass \ Y_term r).

Lemma incl_Y'_Ypass : ∀ r, next_Y_inner_loop r ⊆ Y_pass.
Proof.
  intro r. apply incl_tran with (Y_pass \ Y_term r).
  - apply filter_incl.
  - apply incl_sdiff.
Qed.

Definition inner_loop_iter (r : in_iter_data) (tij : basic_test) : in_iter_data :=
  match Y r with
  | ∅ => r
  | _ :: _ => let Y' := next_Y_inner_loop r in
              match Y' with
              | ∅ => {| Y := Y';
                        Y_term := Y_term r;
                        tt := tt r |}
              | _ :: _ => {| Y := Y';
                             Y_term := Y_term r ∪ Y';
                             tt := tij :: tt r |}
              end
  end.

Lemma inner_loop_iter_invar r :
  in_data_invariant r -> ∀ tij, in_data_invariant (inner_loop_iter r tij).
Proof.
  intros iv *. unfold inner_loop_iter.
  destruct (Y r) as [| _ _].
  - exact iv.
  - destruct iv as [InInv1 InInv2 ndY ndYterm].
    case_eq (next_Y_inner_loop r); [ intros e | intros nn snn e; rewrite <- e; clear nn snn e].
    + split; cbn.
      * apply empty_incl.
      * exact InInv2.
      * apply NoDup_empty.
      * exact ndYterm.
    + assert (ndY' : NoDup (next_Y_inner_loop r)).
      { unfold next_Y_inner_loop. apply filter_nodup.
        apply sdiff_nodup; [ exact ndYpass | exact ndYterm]. }
      split; cbn.
      * apply incl_Y'_Ypass.
      * apply union_incl; [exact InInv2 |idtac]. apply incl_Y'_Ypass.
      * exact ndY'.
      * apply union_nodup; [exact ndYterm | idtac]. exact ndY'.
Qed.

Definition size (l1 l2 : set ℕ²) : ℕ :=
  match l1 with
  | nil => O
  | _ :: _ => S (length l2)
  end.

Definition size_of_data r := size (Y r) (Y_pass \ Y_term r).

Lemma dec_size :
  ∀ (r : in_iter_data),
  in_data_invariant r ->
  ∀ (tij : basic_test) n l,
  Y r = n :: l -> (size_of_data (inner_loop_iter r tij) < size_of_data r)%nat.
Proof.
  intros * [invr1 invr2 nd1 nd2] * e. unfold inner_loop_iter. rewrite e.
  case_eq (next_Y_inner_loop r).
  - intros e'. unfold size_of_data. simpl. rewrite e. simpl. apply Nat.lt_0_succ.
  - intros n' l' e'.
    (* unfold size_of_data. Opaque union. simpl. Transparent union. *)
    change ((S (length (Y_pass \ (n' ⊕ (Y_term r ∪ l'))))
             < size (Y r) (Y_pass \ Y_term r))%nat).
    rewrite e. apply lt_n_S.
    assert (nd : NoDup ((Y_pass) \ (n' ⊕ (Y_term r ∪ l')))).
    { apply sdiff_nodup.
      - assumption.
      - apply sadd_nodup. apply union_nodup.
        + assumption.
        + assert (ndY' : NoDup (next_Y_inner_loop r)).
          { unfold next_Y_inner_loop. apply filter_nodup. apply sdiff_nodup; assumption. }
          rewrite e' in ndY'. destruct (NoDup_cons_elim ndY'); assumption.
    }
    apply (sincl_len_lt nd). apply sincl_sdiff.
    + intros x ix. destruct (sadd_elim ix) as [e_x | ix'].
      * rewrite e_x. apply (incl_Y'_Ypass r). rewrite e'; left; reflexivity.
      * destruct (union_elim ix') as [ix'1 | ix'2].
        { apply invr2. assumption. }
        { apply (incl_Y'_Ypass r). rewrite  e'. right. assumption. }
    + assert (in1 : (n' :: l') ⊆ Y_pass \ (Y_term r))
       by (rewrite <- e'; apply filter_incl).
     split.
     * intros x ix. apply sadd_intro2. apply union_intro1. assumption.
     * exists n'.
       { apply sadd_intro1. reflexivity. }
       { assert (in' : n' ∈ Y_pass \ Y_term r).
         - apply in1. left. reflexivity.
         - destruct (sdiff_elim in'); assumption. }
Qed.

Unset Implicit Arguments.
Function in_loop (r : in_iter_data) (di : in_data_invariant r) (tij : basic_test)
  {measure size_of_data r} : in_iter_data :=
  match Y r with
   | ∅ => r
   | _ :: _ => in_loop (inner_loop_iter r tij) (inner_loop_iter_invar di tij) tij
  end.
Proof.
  intros * di *. apply dec_size. assumption.
Defined.

(* defines :
in_loop_tcc
in_loop_terminate
in_loop_ind
in_loop_rec
in_loop_rect
R_in_loop_correct
R_in_loop_complete
 *)

(*----------- Correctness of inner loop --------------------------*)

(* For invariant (c) *)
Lemma inner_loop_iter_invar_c {k r bt} :
  0 < Pr k (tt r) -> (∃ t, tt r = bt :: t) -> Pr k (tt (inner_loop_iter r bt)) > 0.
Proof.
  intros gt0 i. unfold inner_loop_iter.
  destruct (Y r) as [ | _ _].
  { assumption. }
  case_eq (next_Y_inner_loop r); simpl.
  - intro; exact gt0.
  - intros n l e. apply mult_strict_pos_intro.
    + destruct i as [x ettr]. rewrite ettr in gt0. rewrite Pr_cons in gt0.
      apply (mult_strict_pos_elim1 _ (Pr k x)).
      * apply not_neg_t.
      * exact gt0.
    + exact gt0.
Qed.

Lemma inner_loop_iter_invar_c2 {k r} :
  Pr k (tt r) == 0 -> ∀ bt, Pr k (tt (inner_loop_iter r bt)) == 0.
Proof.
  intros e bt. unfold inner_loop_iter.
  destruct (Y r) as [ | _ _].
  - exact e.
  - destruct (next_Y_inner_loop r) as [ | n l]; simpl.
    + exact e.
    + rewrite e. apply Qmult_0_r.
Qed.

Lemma in_loop_invar_c {k r bt di} :
  0 < Pr k (tt r) -> (∃ t, (tt r) = bt :: t) -> 0 < Pr k (tt (in_loop r di bt)).
Proof.
  intros gt0 i. functional induction in_loop r di bt.
(* For memory: equivalent to
       revert gt0 i; pattern r, di, bt, (in_loop r di bt); apply in_loop_ind; clear r di bt;
       intros r di tij ; [ intros e gt0 i | intros _x _x0 e IHi0 gt0 i ].
*)
 - exact gt0.
 - apply IHi0.
   + apply (inner_loop_iter_invar_c gt0 i).
   + destruct i as [t ettr]. unfold inner_loop_iter.
     destruct (Y r).
     * discriminate e.
     * destruct (next_Y_inner_loop r); simpl.
       { exists t; exact ettr. }
       { exists (tt r); reflexivity. }
Qed.

Lemma in_loop_invar_c2 {k r di} :
  Pr k (tt r) == 0 -> ∀ bt, Pr k (tt (in_loop r di bt)) == 0.
Proof.
  intros e. intro bt. functional induction in_loop r di bt.
  - exact e.
  - apply IHi. apply (inner_loop_iter_invar_c2 e tij).
Qed.

(* For invariant (d) *)

Lemma in_loop_nextY_n r bt :
  Y r <> ∅ ->  next_Y_inner_loop r = Y (inner_loop_iter r bt).
Proof.
  intros ne. unfold inner_loop_iter.
  destruct (Y r) as [| nn lnn].
  - case ne; reflexivity.
  - destruct (next_Y_inner_loop r); reflexivity.
Qed.


Lemma stable r bt :
  next_Y_inner_loop r = ∅ -> let r' := inner_loop_iter r bt in
  Y_term r = Y_term r' ∧ tt r =  tt r'.
Proof.
  intros e r'. unfold r'; unfold inner_loop_iter.
  destruct (Y r).
  - repeat split; reflexivity.
  - rewrite e. simpl. repeat split; reflexivity.
Qed.

Lemma stable_Y r bt :
  next_Y_inner_loop r = ∅ -> next_Y_inner_loop (inner_loop_iter r bt) = ∅.
Proof.
  intros e. destruct (stable r bt e) as [eb ec].
  unfold next_Y_inner_loop in * |- *.
  rewrite eb in e; rewrite ec in e. assumption.
Qed.

(* *)

Definition rditer := Recdef.iter.

Lemma iter_S_right {A : Type} (f : A -> A) n x :
  rditer A (S n) f x = rditer A n f (f x).
Proof.
induction n as [| n IHn]; simpl.
- reflexivity.
- rewrite <- IHn. simpl. reflexivity.
Qed.

Definition loop_body bt r := inner_loop_iter r bt.

Inductive ufp (iid: in_iter_data) r bt : Prop :=
  ufp_intro :
  ∀ n,
    iid = rditer _ n (loop_body bt) r ->
    (Y r <> ∅ -> (0 < n)%nat ∧
                 Y (rditer _ (pred n) (loop_body bt) r) <> ∅ ∧
                 next_Y_inner_loop (rditer _ (pred n) (loop_body bt) r) = ∅) ->
    (Y r = ∅ -> (n = 0)%nat)
    -> ufp iid r bt.

Lemma unfold_fixed_point r (di : in_data_invariant r) bt :
  ufp (in_loop r di bt) r bt.
Proof.
functional induction in_loop r di bt.
- exists 0%nat; simpl.
  + reflexivity.
  + intro ne; case (ne e).
  + reflexivity.
- destruct IHi as [n eil Hnempty Hempty].
  assert (findit : in_loop (inner_loop_iter r tij) (inner_loop_iter_invar di tij) tij =
                   rditer in_iter_data (S n) (loop_body tij) r).
  { rewrite iter_S_right. unfold loop_body at 2. exact eil. }
  exists (S n).
  + exact findit.
  + intros Yne. clear e. split; [auto with arith | ].
    change (pred (S n)) with n.
    assert (choice: {Y (inner_loop_iter r tij) = ∅} + {Y (inner_loop_iter r tij) <> ∅}).
    { apply list_eq_dec. apply eq_dec_nat2. }
    destruct choice as [e0 | ne0].
    * rewrite (Hempty e0). simpl. split.
      { exact Yne. }
      { rewrite <- e0. apply (in_loop_nextY_n _ tij Yne). }
    * destruct (Hnempty ne0) as [pos h1].
      assert (h2: rditer in_iter_data (pred n) (loop_body tij) (inner_loop_iter r tij) =
                  rditer in_iter_data n (loop_body tij) r).
      { change (inner_loop_iter r tij) with (loop_body tij r).
        rewrite <- (Nat.succ_pred_pos n pos).
        rewrite iter_S_right; reflexivity. }
      rewrite <- h2. exact h1.
  + rewrite e; discriminate.
Qed.

Lemma in_loop_nextI r di bt : Y r <> ∅ -> next_Y_inner_loop (in_loop r di bt) = ∅.
Proof.
  intros ne.
  destruct (unfold_fixed_point r di bt) as [n eil Hnempty Hempty].
  rewrite eil. destruct (Hnempty ne) as [posn [H1 empY']].
  rewrite <- (Nat.succ_pred_pos n posn).
  replace (rditer in_iter_data (S (pred n)) (loop_body bt) r) with
      (loop_body bt (rditer in_iter_data (pred n) (loop_body bt) r)).
  - unfold loop_body at 1. apply stable_Y. exact empY'.
  - reflexivity.
Qed.

(* *)

Inductive all_diff_Ypass (r: in_iter_data) : Prop :=
| all_diff_Ypass_intro :
  all_false Y_pass (eq_prob (tt r)) ->
  all_diff_Ypass r.

Inductive all_diff_Yterm (r: in_iter_data) : Prop :=
| all_diff_Yterm_intro :
  all_false (Y_term r) (eq_prob (tt r)) ->
  all_diff_Yterm r.

Lemma Y_empty r :
  next_Y_inner_loop r = ∅ ->
  all_diff_Yterm r ->
  all_diff_Ypass r.
Proof.
  unfold next_Y_inner_loop; unfold filter_eq.
  intros e [ad]. apply all_diff_Ypass_intro. revert e ad.
  apply filter_sdiff_empty.
Qed.

(* *)
Inductive sall_diff_Yterm (r : in_iter_data) (bt : basic_test) : Prop :=
| sall_diff_Yterm_intro :
  (∀ n, all_false (Y_term r) (eq_prob ((repeat bt n) ++ (tt r)))) ->
  sall_diff_Yterm r bt.

Lemma in_loop_iter_test {n r bt t} :
 tt r = repeat  bt n ++ t ->
 tt (inner_loop_iter r bt) = tt r ∨
 tt (inner_loop_iter r bt) = repeat bt (S n) ++ t.
Proof.
  intros e. unfold inner_loop_iter.
  case (Y r).
  - left; reflexivity.
  - intros _ _. destruct (next_Y_inner_loop r); simpl.
    + left. reflexivity.
    + right. rewrite e. reflexivity.
Qed.

(* Towards Ipass_pos *)
Inductive all_diff_Y (r : in_iter_data) (bt : basic_test) : Prop :=
| all_diff_Y_intro :
    (∀ x, x ∈ (next_Y_inner_loop r)  ->  eq_prob_bt bt x = false) ->
    all_diff_Y r bt.

Inductive all_pos_Ypass1 (bt : basic_test) : Prop :=
| all_pos_Ypass1_intro :
  (∀ x1 x2, (x1, x2) ∈ Y_pass -> 0 < Prb x1 bt ∧ 0 < Prb x2 bt) ->
  all_pos_Ypass1 bt.

Inductive all_pos_Ypass2 (t : test) : Prop :=
| all_pos_Ypass2_intro :
  (∀ x1 x2, (x1, x2) ∈ Y_pass -> 0 < Pr x1 t ∧ 0 < Pr x2 t) ->
  all_pos_Ypass2 t.

Inductive Ipass_prop (r : in_iter_data) (bt : basic_test) : Prop :=
| Ipass_prop_intro :
   (∀ x1 x2, (x1, x2) ∈ Y_pass ->
      (exists n t, (tt r = repeat bt n ++ t) ∧ (¬ Pr x1 t == Pr x2 t))
      ∨ eq_prob_bt bt (x1, x2) = false) ->
  Ipass_prop r bt.

Lemma Ipass_pos {r bt} : all_pos_Ypass1 bt -> Ipass_prop r bt -> all_diff_Y r bt.
Proof.
  intros [ap] [ip]. apply all_diff_Y_intro. intros [x1 x2] inx; simpl.
  unfold next_Y_inner_loop in inx. unfold filter_eq in inx.
  destruct (filter_elim inx) as [inx' epx]; clear inx.
  destruct (sdiff_elim inx') as [inx'' _]; clear inx'.
  destruct (ip _ _ inx'') as [[n [l [etr nex]]] | nex]; clear ip; simpl in * |-.
  - rewrite etr in epx. unfold eq_prob in epx; simpl in epx.
    destruct (eq_nat_dec n O) as [e | ne].
    + rewrite e in epx; simpl in epx. case nex. apply (Qeq_bool_eq _ _ epx). 
    + apply (Qeq_bool_false_eq (Qeq_eq_bool_false _ _ nex)); clear etr nex; intro exbt.
      rewrite Qeq_bool_iff in epx.
      do 2 rewrite rep_bt in epx. clear ne.
      rewrite (natpow_eq _ exbt) in epx.
      revert epx; apply Qmult_inj_l. apply Qpos_not_null. apply Qpow_spos.
      destruct (ap _ _ inx''); assumption.
  - assumption.
Qed.

(* Towards in_loop_term *)

Lemma inner_loop_iter_pos2_inv {r bt} :
  all_pos_Ypass1 bt ->
  all_pos_Ypass2 (tt r) ->
  all_pos_Ypass2 (tt (inner_loop_iter r bt)).
Proof.
  intros [ap1] [ap2]. apply all_pos_Ypass2_intro.
  intros x1 x2 i; specialize (ap1 _ _ i); specialize (ap2 _ _ i).
  assert (choice : tt (inner_loop_iter r bt) = tt r ∨
                   tt (inner_loop_iter r bt) = bt :: tt r)
  by (apply (@in_loop_iter_test 0 r bt (tt r)); reflexivity).
  destruct choice as [e|e]; rewrite e.
  - exact ap2.
  - simpl.
    destruct ap2 as [i11 i12]; destruct ap1 as [i21 i22].
    split; apply mult_strict_pos_intro; assumption.
Qed.

Lemma inner_loop_iter_fpass_inv {r bt} :
  Ipass_prop r bt ->
  Ipass_prop (inner_loop_iter r bt) bt.
Proof.
  intros [ip]. apply Ipass_prop_intro. simpl.
  intros x1 x2 i. specialize (ip _ _ i).
  assert (choice : tt (inner_loop_iter r bt) = tt r ∨
                   tt (inner_loop_iter r bt) = bt :: tt r)
  by (apply (@in_loop_iter_test 0 r bt (tt r)); reflexivity).
  destruct choice as [e|e]; rewrite e.
  - apply ip.
  - destruct ip as [[x [lbt [ettr nepr]]] | nepr].
    + left. exists (S x). exists lbt. simpl. split.
      * rewrite ettr; reflexivity.
      * exact nepr.
    + right; exact nepr.
Qed.

Opaque union sprod (* in order to keep set notations in goals *).
Lemma in_loop_iter_term {r bt} :
  all_pos_Ypass1 bt ->
  all_pos_Ypass2 (tt r) ->
  Ipass_prop r bt ->
  sall_diff_Yterm r bt ->
  sall_diff_Yterm (inner_loop_iter r bt) bt.
Proof.
  intros apos1 apos2 pp adiff.
  destruct (Ipass_pos apos1 pp) as [disting]; clear pp.
  unfold inner_loop_iter.
  destruct (Y r) as [ | _ _].
  - exact adiff.
  - match goal with [ |- sall_diff_Yterm ?T bt ] => pose (Y' := T) end; fold Y'.
    apply sall_diff_Yterm_intro. intro n. red.
    destruct adiff as [adiff]. unfold all_false in adiff.
    intros x.
    revert Y'.
    case_eq (next_Y_inner_loop r); [intros e | intros y' l eyl] ; simpl; intro ix.
    + apply (adiff n x ix).
    + rewrite <- eyl in ix. clear y' l eyl. rewrite repeat_app_cons.
      destruct (union_elim ix) as [ix1 | ix2]; clear ix.
      * apply (adiff (S n) _ ix1).
      * clear adiff. specialize (disting _ ix2).
        destruct (filter_elim ix2) as [ixd epx]; clear ix2.
        destruct x as [x1 x2].
        unfold eq_prob in * |- *.
        rewrite Qeq_bool_iff in epx.
        apply (Qeq_bool_false_eq disting); clear disting; intro exbt.
        do 2 rewrite rep_bt in exbt.
        rewrite <- epx in exbt.
        destruct (sdiff_elim ixd) as [ix_pass _nix_term]; clear ixd.
        assert (enp : natpow (Prb x1 bt) (S n) == natpow (Prb x2 bt) (S n)).
        { revert exbt; apply Qmult_inj_r.
          destruct apos2 as [apos2]. destruct (apos2 _ _ ix_pass) as [pos1 pos2].
          apply Qpos_not_null; exact pos1. }
        destruct apos1 as [apos1]. destruct (apos1 _ _ ix_pass) as [pos1 pos2].
        apply Qlt_le_weak in pos1. apply Qlt_le_weak in pos2.
        exact (pow_root n _ _ pos1 pos2 enp).
Qed.
Transparent union sprod.

Lemma in_loop_term {r bt t} di :
  (∃ n, (tt r) = (repeat bt n) ++ t) ->
  all_pos_Ypass1 bt ->
  all_pos_Ypass2 (tt r) ->
  Ipass_prop r bt ->
  sall_diff_Yterm r bt ->
  sall_diff_Yterm (in_loop r di bt) bt.
Proof.
  intros n_ettr apos1 apos2 pp adiff.
  functional induction in_loop r di bt as [r di bt e | r di bt _x _l e IHi].
  - exact adiff.
  - apply IHi; clear IHi.
    + destruct n_ettr as [n ettr].
      destruct (in_loop_iter_test ettr) as [etti | etti]; rewrite etti.
      * exists n. apply ettr.
      * exists (S n). reflexivity.
    + apply (apos1).
    + apply (inner_loop_iter_pos2_inv apos1 apos2).
    + apply (inner_loop_iter_fpass_inv pp).
    + apply (in_loop_iter_term apos1 apos2 pp adiff).
Qed.

(* *)

(* Invariance of Y_pass *)

Lemma in_loop_pass {r bt} t di :
  (∃ n, (tt r) = (repeat bt n) ++ t) ->
  all_pos_Ypass1 bt ->
  all_pos_Ypass2 (tt r) ->
  Ipass_prop r bt ->
  sall_diff_Yterm r bt ->
  Y r <> ∅ ->
  all_diff_Ypass (in_loop r di bt).
Proof.
  intros rp p2 pp fp df nemp. apply Y_empty.
  - apply (in_loop_nextI _ _ _ nemp).
  - destruct (in_loop_term di rp p2 pp fp df) as [af].
    apply all_diff_Yterm_intro. apply (af O).
Qed.

End sec_with_Y_pass.

(* ---------------------------------------------------------------------- *)
(* Outer loop *)

Record out_iter_data : Type := mk_out_iter_data {
  I' : set ℕ;
  Y_rem : set ℕ²;
  Y_pass : set ℕ²;
  et : test; (* enhanced test *)
}.


Inductive out_data_invariant0 (r: out_iter_data) : Prop :=
| out_data_invariant0_intro :
    (∀ i j, (i,j) ∈ (Y_rem r) -> i <> j) ->
    NoDup (I' r) -> NoDup (Y_rem r) -> NoDup (Y_pass r) ->
    Y_rem r ⊆ I0×I0 ->
    out_data_invariant0 r.

Definition outer_loop_iter1 (r : out_iter_data) (bt : basic_test) : out_iter_data :=
  let nI' := filter_pos bt (I' r) in
  let Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI')) in
  {| I' := nI';
     Y_rem := (Y_rem r ∩ (nI' × nI')) \ Ydis;
     Y_pass := (Y_pass r ∩ (nI' × nI')) ∪ Ydis;
     et := bt :: et r |}.

(* Definition of outer loop and the whole algorithm *)

Definition mk_in_from_out (r : out_iter_data) (bt : basic_test) : in_iter_data :=
  let rnew := outer_loop_iter1 r bt in
  {| Y := Y_pass rnew;
     Y_term := ∅;
     tt := et rnew |}.

Lemma nd_Y_pass r (dio : out_data_invariant0 r) bt :
  let rnew := outer_loop_iter1 r bt in
  NoDup (Y_pass rnew).
Proof.
  intros.
  destruct dio as [OutI1 OutI2 OutI3 OutI4 OutI5].
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  unfold outer_loop_iter1 in rnew; fold nI' Ydis in rnew. apply union_nodup.
    + apply inter_nodup_left. exact OutI4.
    + apply filter_nodup. apply inter_nodup_left. exact OutI3.
Qed.

Lemma out_data_in_data_invar r (dio : out_data_invariant0 r) bt :
  let rnew := outer_loop_iter1 r bt in
  in_data_invariant (Y_pass rnew) (mk_in_from_out r bt).
Proof.
  intros.
  assert (ndY : NoDup (Y_pass rnew)) by apply (nd_Y_pass r dio).
  destruct dio as [OutI1 OutI2 OutI3 OutI4 OutI5].
  unfold mk_in_from_out. split; unfold Y, Y_term, tt; fold rnew.
  - apply incl_refl.
  - apply empty_incl.
  - apply ndY.
  - apply NoDup_empty.
Qed.

Definition outer_loop_iter2 r (dio : out_data_invariant0 r) bt : out_iter_data :=
  let rnew := outer_loop_iter1 r bt in
  let rin := mk_in_from_out r bt in
  let ndYp := nd_Y_pass r dio bt in 
  let rin' := in_loop (Y_pass rnew) ndYp rin (out_data_in_data_invar r dio bt) bt in
  {| I' := I' rnew;
     Y_rem := Y_rem rnew;
     Y_pass := Y_pass rnew;
     et := tt rin' |}.

Ltac clrin' :=
  match goal with [ rin' := in_loop _ _ _ _ _ |- _ ] => clear rin' end.
Ltac clrin :=
  match goal with [ rin := mk_in_from_out _ _ |- _ ] => clear rin end.
Ltac clnew :=
  match goal with [ rnew := outer_loop_iter1 _ _ |- _ ] => clear rnew end.

Ltac cloli2 := clrin'; clrin; clnew.

Lemma outer_loop_iter_invar0 :
  ∀ r (dio : out_data_invariant0 r) bt,
    out_data_invariant0 (outer_loop_iter2 r dio bt).
Proof.
  intros *.
  unfold outer_loop_iter2.
  pose (rnew := outer_loop_iter1 r bt); fold rnew.
  pose (rin := mk_in_from_out r bt); fold rin.
  pose (ndYp := nd_Y_pass r dio bt); fold ndYp.
  pose (rin' := in_loop (Y_pass rnew) ndYp rin (out_data_in_data_invar r dio bt) bt); fold rin'.
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  destruct dio as [OutI1 OutI2 OutI3 OutI4 OutI5].
  assert (ndYdis : NoDup Ydis).
  { cloli2. apply filter_nodup. apply inter_nodup_left. exact OutI3. }
  assert (ndY_pass : NoDup (Y_pass r ∩ (nI' × nI') ∪ Ydis)).
  { cloli2. apply union_nodup.
    - apply inter_nodup_left. exact OutI4.
    - exact ndYdis. }
  assert (nd_nI' : NoDup nI'). { apply filter_nodup. exact OutI2. }
  unfold outer_loop_iter1 in rnew; fold nI' Ydis in rnew.
  case_eq (Y_rem rnew).
  - intros e. split; cbn.
    + intros * F; case F.
    + exact nd_nI'.
    + apply NoDup_empty.
    + exact ndY_pass.
    + apply empty_incl.
  - intros n3 l e. rewrite <- e. split; cbn.
    + intros i j inn. apply OutI1.
      destruct (sdiff_elim inn) as [inn1 _]; clear inn.
      destruct (inter_elim inn1) as [inn _]; clear inn1.
      exact inn.
    + exact nd_nI'.
    + apply sdiff_nodup.
      * apply inter_nodup_left. exact OutI3.
      * exact ndYdis.
    + exact ndY_pass.
    + eapply incl_tran.
      { apply incl_sdiff. }
      eapply incl_tran.
      { apply inter_incl1. }
      apply OutI5.
Qed.

Definition size_of_out_data x := length (Y_rem x).

Lemma hd_in {A : Type} {x:A} {l u} : x :: l ⊆ u -> x ∈ u.
Proof.
  intro xlu. apply xlu. left. reflexivity.
Qed.

Lemma inI0 {r} (di : out_data_invariant0 r) : Y_rem r ⊆ I0×I0.
Proof.
  destruct di; assumption.
Qed.

(*
Definition out_loop r (di : out_data_invariant0 r) : out_iter_data.
  refine (
  match Y_rem r as x return (x ⊆ I0 × I0 -> out_iter_data) with
  | nil => fun _ => r
  | (i , j) :: _ => fun YI0 => let (bt, _) := oracle i j (hd_in YI0) in _
  end (inI0 di)).
 *)

(* Bug : Coq  should allow this
Function out_loop r (di : out_data_invariant0 r)
                    {measure size_of_out_data r} : out_iter_data :=
  match Y_rem r as x return (x ⊆ I0×I0 -> _) with
  | nil => fun _ => r
  | (i , j) :: _ => fun YI0 => let (bt, _) := oracle i j (hd_in YI0) in
                    out_loop (outer_loop_iter2 r di bt) (outer_loop_iter_invar0 r di bt)
  end (inI0 di).
*) 

Inductive resu_pick r : Set :=
| P_empty : Y_rem r = ∅ -> resu_pick r
| P_nonempty : ∀ i j s,
    Y_rem r = (i, j) :: s -> distinguish Prb i j -> resu_pick r.

Definition pick {r} (di : out_data_invariant0 r) : resu_pick r.
  refine (
   match Y_rem r as x return (Y_rem r = x -> _) with
   | ∅ => P_empty r
   | (i,j) :: s => fun e => P_nonempty r i j s e (oracle i j _)
   end eq_refl).
  apply (inI0 di). rewrite e. left; reflexivity.
Defined.

Function out_loop r (di : out_data_invariant0 r)
         {measure size_of_out_data r} : out_iter_data :=
  match pick di with
  | P_empty _ e => r
  | P_nonempty _ i j s e dis => let (bt, _) := dis in
                                out_loop (outer_loop_iter2 r di bt)
                                         (outer_loop_iter_invar0 r di bt)
  end.
  intros r di i j l teq ordist bt disting_ij _ _. cbn. unfold size_of_out_data.
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  fold nI'. fold Ydis.
  destruct di as [OutI1 OutI2 OutI3 OutI4 Out5].
  assert (nd : NoDup (Y_rem r ∩ (nI' × nI') \ Ydis)).
  { apply sdiff_nodup.
    - apply inter_nodup_left. exact OutI3.
    - apply filter_nodup. apply inter_nodup_left. exact OutI3. }
  apply (sincl_len_lt nd). split.
  - apply incl_tran with (Y_rem r ∩ (nI' × nI')).
    + apply incl_sdiff.
    + apply inter_incl1.
  - exists (i,j).
    + rewrite teq. left; reflexivity.
    + intro iij. destruct (sdiff_elim iij) as [iij1 iij2]. apply iij2.
      apply filter_intro.
      * exact iij1.
      * apply negb_true_iff. apply Qeq_eq_bool_false.
        apply disting_ij. apply OutI1. rewrite teq; apply in_eq.
Qed.


(* In Type instead of Set, unless I'f : set ℕ is replaced with list ℕ *)
Record final_data : Type := mk_final_data {
  I'f : set ℕ;
  etf : test;
}.

(* The algorithm for computing enhanced tests *)
(* m is the number of equivalence classes *)

(* initilization: I, I*I|_ i<j, ∅, ω *)
Definition initial_I : set ℕ := nodup I0.
Definition initial_Y : set ℕ² := filter_lt (nodup (I0 × I0)).

Definition init_data : out_iter_data :=
  {| I' := initial_I;
     Y_rem := initial_Y;
     Y_pass := ∅;
     et := ∅ |}.

Lemma initial_Y_lt {i j} : (i,j) ∈ initial_Y -> (i < j)%nat.
Proof. unfold initial_Y. apply filter_lt_list. Qed.

Lemma init_data_invariant0 : out_data_invariant0 init_data.
Proof.
  unfold init_data; split; cbn.
  - intros i j ijY. generalize (initial_Y_lt ijY). intuition.
  - apply NoDup_nodup.
  - unfold initial_Y. apply filter_nodup. apply NoDup_nodup.
  - apply NoDup_empty.
  - unfold initial_Y. eapply incl_tran. apply filter_incl.
    intros a an. apply (nodup_In_elim an).
Qed.

Definition algo_compt_enhanced_test : final_data :=
  let r := init_data in
  let final_r := out_loop r init_data_invariant0 in
  {| I'f := I' final_r;
     etf := et final_r |}.

(* Termination of the algorithm is thus guaranteed. *)

(*---------------------Correctness--------------------------*)

(*Invariants of the outer loop*)

(* *)
(*
Ltac unfold_outer_loop_iter2 :=
  Opaque union sprod; unfold outer_loop_iter2; cbn; Transparent union sprod.
 *)

(* Invariant (a): Y_pass U Y_rem = {(i,j) ∈ I' X I' | i < j} *)
Inductive out_data_invariant_a (d: out_iter_data) : Prop :=
| out_data_invariant_a_intro :
    (Y_rem d ∪ Y_pass d) ⊆ (filter_lt (I' d × I' d)) ->
    (filter_lt (I' d × I' d)) ⊆ (Y_rem d ∪ Y_pass d) ->
    out_data_invariant_a d.

(* Invariant (a) is preserved after one iteration of the outer loop. *)
Opaque union sprod (* in order to keep set notations in goals *).
Lemma outer_loop_iter_invar_a {r dio bt} :
  out_data_invariant_a r -> out_data_invariant_a (outer_loop_iter2 r dio bt).
Proof.
  intros [odi_a_1 odi_a_2].
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  pose (nI'2 := nI' × nI').
  apply out_data_invariant_a_intro.
  - unfold outer_loop_iter2; cbn; fold nI'; fold Ydis; fold nI'2.
    apply incl_tran with (Y_rem r ∩ nI'2 ∪ (Y_pass r ∩ nI'2)).
    + apply union_sdiff2. apply filter_incl.
    + eapply incl_tran; [apply union_inter_distrib | ].
      eapply incl_tran; [apply (incl_inter_monot nI'2 odi_a_1) | ].
      intros x H; destruct (inter_elim H) as [xfilt xfipos]; clear H.
      destruct (filter_elim xfilt) as [_ lx]; clear xfilt.
      apply filter_intro; [exact xfipos | exact lx].
  - unfold outer_loop_iter2; cbn; fold nI'; fold Ydis; fold nI'2.
    eapply incl_tran; [ | apply union_sdiff].
    eapply incl_tran; [ | apply union_inter_factor].
    eapply incl_tran; [ | apply (incl_inter_monot nI'2 odi_a_2)].
    intros x H; destruct (filter_elim H) as [xfi lx]; clear H.
    apply inter_intro; [ | exact xfi].
    apply filter_intro.
    + assert(fiic: nI'2 ⊆ (I' r × I' r)).
      { apply prod_monot; apply filter_incl. }
      apply fiic. exact xfi.
    + exact lx.
Qed.
Transparent union sprod.

(* Invariant (b): I' is nonempty. *)
Inductive out_data_invariant_b (d: out_iter_data) : Prop :=
| out_data_invariant_b_intro : I' d <> ∅ -> out_data_invariant_b d.

(* Invariant (b) is preserved after one iteration of the outer loop. *)
Lemma outer_loop_iter_invar_b {r i j dio bt} :
  out_data_invariant_a r ->
  (i,j) ∈ (Y_rem r) -> (i <> j -> ¬ Prb i bt == Prb j bt) ->
  out_data_invariant_b (outer_loop_iter2 r dio bt).
Proof.
  intros [OutI2_1 OutI2_2] ij dist_bt.
  apply out_data_invariant_b_intro.
  unfold outer_loop_iter2; cbn.
  assert (ijI' : i ∈ I' r ∧ j ∈ I' r).
  { apply in_prod_elim. eapply filter_incl. apply OutI2_1. apply incl_union1. apply ij. }
  destruct ijI' as [iI' jI'].
  destruct dio as [OutI1 OutI2 OutI3 OutI4].
  assert (Prij_pos: 0 < Prb i bt ∨ 0 < Prb j bt).
  { apply pos_neq; try apply not_neg_bt. apply Qeq_eq_bool_false.
    apply dist_bt; apply OutI1. exact ij. }
  unfold filter_pos. apply exists_nonempty.
  destruct Prij_pos as [li | lj]; [exists i | exists j]; apply filter_intro; try assumption.
   - apply pos_0. exact li.
   - apply pos_0. exact lj.
Qed.

(* Invariant (c): if k ∈ I' then Pr(C_k, t) > 0. *)
Inductive out_data_invariant_c (d: out_iter_data)(I : set ℕ) : Prop :=
| out_data_invariant_c_intro :
    (∀ k, k ∈ I' d -> 0 < Pr k (et d)) ->
    (∀ k, k ∈ I \ I' d  -> Pr k (et d) == 0) ->
    out_data_invariant_c d I.

(* Invariant (c) is preserved after one iteration of the outer loop. *)
Lemma outer_loop_iter_invar_c {r bt dio I} :
  out_data_invariant_c r I -> out_data_invariant_c (outer_loop_iter2 r dio bt) I.
Proof.
  intros [OutI4_1 OutI4_2].
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  apply out_data_invariant_c_intro.
  - unfold outer_loop_iter2; cbn. fold nI'; fold Ydis.
    intros k kposI'.
    apply in_loop_invar_c; simpl.
    + destruct (filter_elim kposI') as [kI' kpos].
      apply mult_strict_pos_intro.
      * apply pos_0_true. exact kpos.
      * exact (OutI4_1 _ kI').
    + exists (et r); reflexivity.
  - unfold outer_loop_iter2; cbn. fold nI'; fold Ydis.
    intros k kdiff.
    apply in_loop_invar_c2; simpl. clear OutI4_1 oracle dio.
    destruct (diff_filter_elim kdiff) as [aa1 | aa2]; clear kdiff.
    + rewrite (OutI4_2 k aa1). apply Qmult_0_r.
    + assert (e0 : Prb k bt == 0).
      { apply Qle_antisym.
        - apply pos_0_false. exact aa2.
        - apply not_neg_bt. }
      rewrite e0. apply Qmult_0_l.
Qed.

(* Invariant (d): ∀ (i,j) ∈ Y_pass, Pr(C_i, t) <> Pr(C_j, t). *)
Inductive out_data_invariant_d (d: out_iter_data) : Prop :=
| out_data_invariant_d_intro :
    all_false (Y_pass d) (eq_prob (et d)) -> out_data_invariant_d d.



(* Additional auxiliary lemmas for outer_loop_iter_invar_d *)

Lemma out_to_in {r} bt dio : ∃ di,
  let rnew := outer_loop_iter1 r bt in
  let ndYp := nd_Y_pass r dio bt in
  let rin := mk_in_from_out r bt in
  let rin' := in_loop (Y_pass rnew) ndYp rin di bt in
  tt rin' = et (outer_loop_iter2 r dio bt).
Proof.
  eexists. simpl. reflexivity.
Qed.

Lemma next_I'_pos {x1 x2 r bt} :
  let r' := filter_pos bt (I' r) in
  (x1, x2) ∈ (r' × r') ->
  0 < Prb x1 bt ∧ 0 < Prb x2 bt.
Proof.
  intros r' ix. unfold r' in ix.
  destruct (in_prod_elim ix) as [ix1 ix2]; clear ix.
  split; apply pos_0_true.
  - destruct (filter_elim ix1) as [_ ip]; exact ip.
  - destruct (filter_elim ix2) as [_ ip]; exact ip.
Qed.

Lemma nextIpass_posbt r x1 x2 bt :
  (x1, x2) ∈ Y_pass (outer_loop_iter1 r bt) ->
  0 < Prb x1 bt ∧ 0 < Prb x2 bt.
Proof.
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  unfold outer_loop_iter1; fold nI' Ydis.
  intro ix.
  destruct (union_elim ix) as [ix'|ix']; clear ix; rename ix' into ix.
  - destruct (inter_elim ix) as [_ ix2]; clear ix.
    apply (next_I'_pos ix2).
  - destruct (filter_elim ix) as [ix1 _]; clear ix.
    destruct (inter_elim ix1) as [_ ix2]; clear ix1.
    apply (next_I'_pos ix2).
Qed.

Lemma nextIpass_post r x1 x2 bt I :
  out_data_invariant_c r I ->
  (x1, x2) ∈ (Y_pass (outer_loop_iter1 r bt)) ->
  0 < Pr x1 (et r) ∧ 0 < Pr x2 (et r). 
Proof.
  pose (nI' := filter_pos bt (I' r)).
  pose (Ydis := filter_neq bt (Y_rem r ∩ (nI' × nI'))).
  unfold outer_loop_iter1; fold nI' Ydis.
  intros [posI' _] ix.
  assert (ixnI'2 : ((x1, x2) ∈ (nI' × nI'))).
  { destruct (union_elim ix) as [ix'|ix']; clear ix.
    - destruct (inter_elim ix') as [_ ix]. exact ix.
    - destruct (filter_elim ix') as [ix _]; clear ix'.
      destruct (inter_elim ix) as [_ ix']. exact ix'. }
  destruct (in_prod_elim ixnI'2) as [ir1 ir2]. split; apply posI'.
  - destruct (filter_elim ir1) as [ir1' _]; exact ir1'.
  - destruct (filter_elim ir2) as [ir2' _]; exact ir2'.
Qed.

(* End of auxiliary lemmas for outer_loop_iter_invar_d *)

(* Invariant (d) is preserved after one iteration of the outer loop. *)
Lemma outer_loop_iter_invar_d {r bt dio I} :
  out_data_invariant_c r I -> out_data_invariant_d r ->
  out_data_invariant_d (outer_loop_iter2 r dio bt).
Proof.
  intros inv4 inv5. apply out_data_invariant_d_intro.
  intro x.
  destruct (out_to_in bt dio) as [idi ett].
  rewrite <- ett; clear ett. revert idi.
  pose (rnew := outer_loop_iter1 r bt); fold rnew.
  pose (rin := mk_in_from_out r bt); fold rin.
  pose (Yp := Y_pass rnew); fold Yp.
  pose (ndYp := nd_Y_pass r dio bt: NoDup Yp); fold ndYp.
  intro idi.
  case_eq (Y_pass (outer_loop_iter2 r dio bt)). 
  - intros ip ix. case ix.
  - intros nn nnl ennl. rewrite <- ennl.
    apply (in_loop_pass _ _ (et r) idi); clear x.
    + exists 1%nat. cbn. reflexivity.
    + apply all_pos_Ypass1_intro. intros x1 x2. simpl.
      intro ix. apply (nextIpass_posbt r x1 x2). exact ix.
    + apply all_pos_Ypass2_intro. simpl. intros x1 x2 ix.
      assert(po1 : 0 < Prb x1 bt ∧ 0 < Prb x2 bt).
      { apply (nextIpass_posbt r). cbn. exact ix. }
      assert(po2 : 0 < Pr x1 (et r) ∧ 0 < Pr x2 (et r)).
      { apply (nextIpass_post r x1 x2 bt I inv4). cbn. exact ix. }
      destruct po1 as [p11 p12]; destruct po2 as [p21 p22].
      split; apply mult_strict_pos_intro; assumption.
    + apply Ipass_prop_intro. cbn. intros x1 x2 ir.
      destruct (union_elim ir) as [ix | ix]; clear ir.
      * left. exists 1%nat. exists (et r).
        { split.
          - reflexivity.
          - destruct (inter_elim ix) as [ix' _]; clear ix.
            destruct inv5 as [inv5]. red in inv5.
            apply Qeq_bool_neq. apply (inv5 _ ix'). }
      * right. destruct (filter_elim ix) as [_ ntrue]; clear ix.
        apply negb_true_iff. exact ntrue.
    + apply sall_diff_Yterm_intro. simpl. intro n.
      red. intros x ix; case ix.
    + simpl; simpl in ennl. rewrite ennl. discriminate.
Qed.

(* Remark: out_data_invariant0 is handled separately *)

Inductive out_data_global_invariant (I : set ℕ) (r: out_iter_data) : Prop :=
  odgi_intro :
    out_data_invariant_a r ->
    out_data_invariant_b r ->
    out_data_invariant_c r I ->
    out_data_invariant_d r ->
    out_data_global_invariant I r.


(* The initial data satisfy all the necessary invariants *)
Lemma init_data_invariant_all : out_data_global_invariant initial_I init_data.
Proof.
  split.
  - apply out_data_invariant_a_intro; simpl.
    + unfold initial_Y. unfold filter_lt. red; intros x ix.
      destruct (filter_elim ix) as [ix' Hl]; clear ix.
      apply filter_intro.
      * apply nodup_In_elim in ix'. destruct x as (x1, x2).
        destruct (in_prod_elim ix') as [x1_I0 x2_I0].
        apply in_prod_intro; apply nodup_In_intro; assumption.
      * exact Hl.
    + (* eapply incl_tran; [ idtac | apply incl_union1]. *)
      unfold initial_Y. intros x ix.
      destruct (filter_elim ix) as [ix' Hl]; clear ix.
      apply filter_intro.
      * apply nodup_In_intro. destruct x as (x1, x2).
        destruct (in_prod_elim ix') as [x1_I0 x2_I0].
        apply in_prod_intro; apply nodup_In_elim; assumption.
      * exact Hl.
  - (*intro pos_m.*) apply out_data_invariant_b_intro; simpl.
    intro em. apply initially_non_empty. apply (nodup_empty I0 em).
  - apply out_data_invariant_c_intro; simpl.
    + intros; red; simpl. repeat constructor.
    + intros k ik. destruct (sdiff_elim ik) as [i ni]; case (ni i).
  - apply out_data_invariant_d_intro; simpl.
    intros x ix; case ix.
Qed.

Lemma out_loop_invar I r (di : out_data_invariant0 r) :
  out_data_global_invariant I r -> out_data_global_invariant I (out_loop r di).
Proof.
  intro odgi_r.
  functional induction out_loop r di.
  - exact odgi_r.
  - apply IHo; clear IHo.
    destruct odgi_r as [odi_a _ odi_c odi_d]. split.
    + apply (outer_loop_iter_invar_a odi_a).
    + assert (irem : (i,j) ∈ Y_rem r) by (rewrite e; left; reflexivity).
      apply (outer_loop_iter_invar_b odi_a irem). assumption (* distinguish Prb i j*).
    + apply (outer_loop_iter_invar_c odi_c).
    + apply (outer_loop_iter_invar_d odi_c odi_d).
Qed.

Lemma out_loop_ensures_empty_Yrem r di : Y_rem (out_loop r di) = ∅.
Proof.
  functional induction out_loop r di.
  - assumption.
  - assumption.
Qed.

Theorem correctness: 
  let (fI, ft) := algo_compt_enhanced_test in
  fI <> ∅ ∧
  (∀ k, k ∈ fI -> 0 < Pr k ft) ∧
  (∀ k, k ∈ (I0 \ fI) -> Pr k ft == 0) ∧
  (∀ i j, i ∈ fI -> j ∈ fI -> i <> j -> ¬ Pr i ft == Pr j ft).
Proof.
  destruct (init_data_invariant_all) as [odi_a odi3 odi_c odi_d].
  pose (oid0 := init_data). fold oid0 in odi_a, odi3, odi_c, odi_d.
  assert (odgi_ini : out_data_global_invariant initial_I oid0)
    by (split; assumption).
  clear odi_a odi3 odi3 odi_c odi_d.
  pose (di0 := init_data_invariant0).
  pose (oidf := out_loop oid0 di0).
  assert (odgi_outloop : out_data_global_invariant initial_I oidf)
    by apply (out_loop_invar _ _ _ odgi_ini).
  destruct odgi_outloop as [odi_a odi3 odi_c odi_d].
  unfold algo_compt_enhanced_test. split; fold di0 oid0 oidf.
  - destruct (odi3) as [odi3p]. exact (odi3p).
  - split; [ | split].
    + destruct odi_c as [odi_c _]. exact odi_c.
    + destruct odi_c as [_ odi_c]. intros k kI0.
      apply odi_c.
      apply sdiff_incl_nodup1. apply kI0.
    + destruct odi_d as [odi_d]. red in odi_d.
      destruct odi_a as [_ odi_a_2].
      assert (Yrem_emp : Y_rem oidf = ∅) by (apply out_loop_ensures_empty_Yrem).
      rewrite Yrem_emp in odi_a_2.
      assert (fil_pass : filter_lt (I' oidf × I' oidf) ⊆ Y_pass oidf).
      { apply (incl_tran _ odi_a_2). apply union_incl; [apply empty_incl | intro; trivial]. }
      intros i j iI' jI' nij.
      destruct (not_eq i j nij) as [l | l]; clear nij; apply Nat.ltb_lt in l.
      * apply Qeq_bool_neq. apply (odi_d (i, j)). apply fil_pass.
        apply filter_intro; [apply in_prod_intro; assumption | assumption].
      * apply Qeq_bool_neq. apply neq_sym.
        apply (odi_d (j, i)). apply fil_pass.
        apply filter_intro; [apply in_prod_intro; assumption | assumption].
Qed.

(* END OF PROOF *)

(* KEEP THE FOLLOWING FOR REPORTING TO COQ TEAM.
   Minimal example of bug? of Coq, unable to find the
   correct instance of P.

out_loop_ind
     : ∀ P : ∀ r : out_iter_data, out_data_invariant0 r -> out_iter_data -> Prop,
       (∀ (r : out_iter_data) (di : out_data_invariant0 r), Y_rem r = ∅ -> P r di r) ->
       (∀ (r : out_iter_data) (di : out_data_invariant0 r) (i j : ℕ) (_x : list ℕ ²),
        Y_rem r = (i, j) :: _x ->
        ∀ (bt : basic_test) (_x0 : i <> j -> Qeq (Prb i bt) (Prb j bt) = false),
        oracle i j =
        exist (fun t : basic_test => i <> j -> Qeq (Prb i t) (Prb j t) = false) bt _x0 ->
        P (outer_loop_iter2 r di bt) (outer_loop_iter_invar0 r di bt)
          (out_loop (outer_loop_iter2 r di bt) (outer_loop_iter_invar0 r di bt)) ->
        P r di (out_loop (outer_loop_iter2 r di bt) (outer_loop_iter_invar0 r di bt))) ->
       ∀ (r : out_iter_data) (di : out_data_invariant0 r), P r di (out_loop r di)

(* Simplified version *)
Variable out_loop_ind'
     : ∀ P : ∀ r : out_iter_data, out_data_invariant0 r -> out_iter_data -> Prop,
       ∀ (r : out_iter_data) (di : out_data_invariant0 r), P r di (out_loop r di)
.

 *)

(*
Lemma show_bug_on_functional_induction : 
  I' (out_loop (init_data I0) (init_data_invariant0)) <> ∅.
Proof.
  intro m.
  Fail functional induction out_loop (init_data I0) (init_data_invariant0 ).
  change ((fun (o : out_iter_data)
               (_ : out_data_invariant0 o)
               (o1 : out_iter_data) =>
             I' o1 <> ∅)
          (init_data I0) (init_data_invariant0) (out_loop (init_data I0) (init_data_invariant0))).
  apply out_loop_ind.
Abort.
*)
End sec_probas.

