(* AUTEURS / AUTHORS :                                     *)
(* Yuxin Deng, Jiatong University, Shanghai                *)
(* Jean-François Monin, Verimag, Université Grenoble Alpes *)
(*                                             2010 - 2019 *)

Require Import List.
Require Import ListSet.

Definition eq_dec (A: Type) := forall (x y: A), {x = y}+{x <> y}.

Class EqDec (A : Type) := {Aeq_dec : eq_dec A}.

Definition eq_dec_prod {A B: Type} : eq_dec A -> eq_dec B -> eq_dec (A*B).
intros edA edB. 
intros [a1 b1] [a2 b2].
  destruct (edA a1 a2) as [ea | nea].
  - destruct (edB b1 b2) as [eb | neb].
    + left. rewrite ea; rewrite eb; reflexivity.
    + right. intro neab. apply neb. injection neab. trivial.
  - right. intro neab. apply nea. injection neab. trivial.
Defined.

Generalizable Variables A B.

Definition EDprod `(EDA: EqDec A) `(EDB: EqDec B) : EqDec (A*B).
  constructor. apply eq_dec_prod; apply Aeq_dec.
Defined.

Definition set := ListSet.set.

(* Set relations and notations *)

Notation "x ∈ y" := (set_In x y) (at level 70, no associativity).
Notation "x ∉ y" := (~ x ∈ y) (at level 70, no associativity).
Definition set_incl {A} (u v: set A) := forall a, a ∈ u -> a ∈ v.
Notation "u ⊆ v" := (set_incl u v) (at level 70, no associativity).
(** Constructive version of not-included %\\%*)
Inductive nincl {A} (u v : set A) : Prop :=
  nincl_intro : forall a, a ∈ u -> a ∉ v -> nincl u v.
Notation "x ⊄ y" := (nincl x y) (at level 70, no associativity).
(** Strict inclusion. *)
(** Equivalent to [u ⊆ v /\ exists a, a ∈ v /\ a ∉ u]. %\\%*)
Definition sincl {A} (u v : set A) : Prop := u ⊆ v /\ v ⊄ u.
Notation "u ⊂ v" := (sincl u v) (at level 70, no associativity).


(* Membership is decidable on EqDec A *)
Definition in_dec `{ED: EqDec A} (a : A) u : {a ∈ u} + {a ∉ u} :=
  set_In_dec Aeq_dec a u.

(* Set operations and notations *)

Notation "∅" := nil.

Definition sadd `{ED: EqDec A} a u := set_add Aeq_dec a u.
Notation "a ⊕ u" := (sadd a u) (at level 60, right associativity).

Definition srem `{ED: EqDec A} a u := set_remove Aeq_dec a u.
Notation "a † u" := (srem a u) (at level 60, right associativity).

Definition union `{ED: EqDec A} u v := set_union Aeq_dec u v.
Notation "u ∪ v" := (union u v) (at level 50, left associativity).

Definition inter `{ED: EqDec A} u v := set_inter Aeq_dec u v.
Notation "u ∩ v" := (inter u v) (at level 40, left associativity).

Definition sdiff `{ED: EqDec A} u v := set_diff Aeq_dec u v.
Notation "u \ v" := (sdiff u v) (at level 45, left associativity).

Definition sprod `(u : set A) `(v: set B) := set_prod u v.
Notation "u × v" := (sprod u v) (at level 40, left associativity).

(* Introduction and elimination principles *)

Lemma empty_elim {A} (a:A) : a ∉ ∅.
Proof. exact (@in_nil _ a). Qed.

Lemma sadd_intro1 `{ED: EqDec A} {a b u} : a = b -> a ∈ b ⊕ u.
Proof. apply set_add_intro2. Qed.

Lemma sadd_intro2 `{ED: EqDec A} {a b u} : a ∈ u -> a ∈ b ⊕ u.
Proof. apply set_add_intro1. Qed.

Lemma sadd_elim `{ED: EqDec A} {a b u} : a ∈ b ⊕ u -> a = b \/ a ∈ u .
Proof. apply set_add_elim. Qed.

Lemma inter_intro `{ED: EqDec A} {a u v} : a ∈ u -> a ∈ v -> a ∈ u ∩ v.
Proof. apply set_inter_intro. Qed.

Lemma inter_elim `{ED: EqDec A} {a u v} : a ∈ u ∩ v -> a ∈ u /\ a ∈ v.
Proof. apply set_inter_elim. Qed.

Lemma union_intro1 `{ED: EqDec A} {a u v} : a ∈ u -> a ∈ u ∪ v.
Proof. apply set_union_intro1. Qed.

Lemma union_intro2 `{ED: EqDec A} {a u v} : a ∈ v -> a ∈ u ∪ v.
Proof. apply set_union_intro2. Qed.

Lemma union_elim `{ED: EqDec A} {a u v} : a ∈ u ∪ v -> a ∈ u \/ a ∈ v.
Proof. apply set_union_elim. Qed.

Lemma sdiff_intro `{ED: EqDec A} {a} {u v} : a ∈ u -> a ∉ v -> a ∈ u \ v.
Proof. apply set_diff_intro. Qed.

Lemma sdiff_elim `{ED: EqDec A} {a} {u v} : a ∈ u \ v -> a ∈ u /\ a ∉ v.
Proof. 
  intro i. destruct (set_diff_iff Aeq_dec a u v) as [sdi1 _]. exact (sdi1 i).
Qed.

Lemma in_prod_intro {A B} {a : A} {b : B} {u v} :
  a ∈ u -> b ∈ v -> (a, b) ∈ u × v.
Proof.
  destruct (in_prod_iff u v a b) as [e i]. tauto.
Qed.

Lemma in_prod_elim {A B} {a : A} {b : B} {u v} :
  (a, b) ∈ u × v -> a ∈ u /\ b ∈ v.
Proof.
  destruct (in_prod_iff u v a b) as [e i]. exact e.
Qed.

(* Properties related to inclusion *)

Lemma incl_tran {A} {u : set A} v {w} : u ⊆ v -> v ⊆ w -> u ⊆ w.
Proof. apply @incl_tran. Qed.

Lemma empty_incl {A} (u : set A) : ∅ ⊆ u.
Proof. intros a F. case F. Qed.

(* *)

Lemma inter_incl1 `{ED: EqDec A} u v : (u ∩ v) ⊆ u.
Proof. intros a i. destruct (inter_elim i); assumption. Qed.

Lemma inter_incl2 `{ED: EqDec A} u v : (v ∩ u) ⊆ u.
Proof. intros a i. destruct (inter_elim i); assumption. Qed.

Lemma union_incl `{ED: EqDec A} {u v w} : u ⊆ w -> v ⊆ w -> (u ∪ v) ⊆ w.
Proof.
  intros iuw ivw a iu. destruct (union_elim iu); intuition.
Qed.

Lemma incl_union1 `{ED: EqDec A} u v : u ⊆ (u ∪ v).
Proof. intros a. apply union_intro1. Qed.

Lemma incl_union2 `{ED: EqDec A} u v : v ⊆ (u ∪ v).
Proof. intros a. apply union_intro2. Qed.

(* *)

Lemma incl_union_monot `{ED: EqDec A} {u v} w : u ⊆ v -> (u ∪ w) ⊆ (v ∪ w).
Proof.
  intros ic a i. destruct (union_elim i) as [iu | iw]; clear i.
  - apply set_union_intro1. exact (ic _ iu).
  - apply set_union_intro2. exact iw.
Qed.

Lemma incl_inter_monot `{ED: EqDec A} {u v} w : u ⊆ v -> (u ∩ w) ⊆ (v ∩ w).
Proof.
  intros ic a i. destruct (inter_elim i) as [iu iw]. apply set_inter_intro.
  - apply ic. apply iu.
  - apply iw.
Qed.

Lemma prod_monot {A B} {u1 u2 : set A} {v1 v2 : set B} :
  u1 ⊆ u2 -> v1 ⊆ v2 -> (u1 × v1) ⊆ (u2 × v2) .
Proof.
  intros ic ic' [a b] i.
  destruct (in_prod_elim i) as [ia ib].
  apply in_prod_intro; auto.
Qed.

(* *)

Lemma sdiff_empty `{ED: EqDec A} u : u ⊆ u \ ∅.
Proof.
  red; intros a iu. apply sdiff_intro; [assumption | intro F; exact F].
Qed.

Lemma sdiff_union `{ED: EqDec A} {a u} v :  a ∈ u -> a ∈ v \/ a ∈ u \ v.
Proof.
  intro iu. destruct (in_dec a v) as [iv | niv]; [left | right].
  - exact iv.
  - apply sdiff_intro; assumption.
Qed.

Lemma incl_sdiff `{ED: EqDec A} {u v} : (u \ v) ⊆ u.
Proof.
  intros a i. destruct (sdiff_elim i); assumption.
Qed.

(* *)

Lemma exists_nonempty {A} u : (exists a:A, a ∈ u) -> u <> ∅.
Proof.
  intros[a iu] e. rewrite e in iu. exact iu.
Qed.

(* ------------------------------------------------------------ *)
(* Specific lemmas *)

Lemma union_sdiff `{ED: EqDec A} {u v w} : (u ∪ v) ⊆ (u \ w) ∪ (v ∪ w).
Proof.
  intros a iuv. destruct (in_dec a w) as [iw | niw].
  - apply union_intro2. apply union_intro2. apply iw.
  - destruct (union_elim iuv) as [iu | iv].
    + apply union_intro1. apply sdiff_intro; assumption.
    + apply union_intro2; apply union_intro1; assumption.
Qed.

Lemma union_sdiff2 `{ED: EqDec A} {u w} v : w ⊆ u -> (u \ w) ∪ (v ∪ w) ⊆ (u ∪ v).
Proof.
  intros ic a i. destruct (union_elim i) as [iuw | ivw]; clear i.
  - apply union_intro1. destruct (sdiff_elim iuw) as [iu _]. exact iu.
  - destruct (union_elim ivw) as [iv | iw]; clear ivw.
    + apply union_intro2. exact iv.
    + apply union_intro1. apply ic. exact iw.
Qed.

Lemma union_inter_distrib `{ED: EqDec A} {u v w} : (u ∩ w) ∪ (v ∩ w) ⊆ ((u ∪ v) ∩ w).
Proof.
  intros a i.
  destruct (union_elim i) as [i' | i']; clear i.
  - destruct (inter_elim i') as [i1 i2]; apply set_inter_intro.
    + apply union_intro1; exact i1.
    + exact i2.
  - destruct (inter_elim i') as [i1 i2]; apply set_inter_intro.
    + apply union_intro2; exact i1.
    + exact i2.
Qed.

Lemma union_inter_factor `{ED: EqDec A} {u v w} : ((u ∪ v) ∩ w) ⊆ (u ∩ w) ∪ (v ∩ w).
Proof.
  intros a i. destruct (inter_elim i) as [iuv iw]; clear i.
  destruct (union_elim iuv) as [iu | iv];
  [apply union_intro1 | apply union_intro2];
  apply inter_intro; assumption.
Qed.

(* ------------------------------------------------------------ *)
(* Negation of inclusion *)

Lemma nincl_not_incl {A} (u v : set A) : u ⊄ v -> ~ u ⊆ v.
Proof.
  intros [a iu niv] i. apply niv. apply i. apply iu.
Qed.

Lemma not_incl_nincl `{ED: EqDec A} {u v : set A} : ~ u ⊆ v -> u ⊄ v.
Proof.
  induction u as [ | a u IHu]; simpl; intro n.
  - case n. apply empty_incl.
  - destruct (in_dec a v) as [iav | niav].
    + destruct IHu as [b ibu nibv].
      * intro ic. apply n. apply incl_cons; [exact iav | exact ic].
      * exists b; [right; exact ibu | exact nibv].
    + exists a; [left; reflexivity | exact niav].
Qed.

Lemma contra_not_incl `{ED: EqDec A} {u v} :
  ~ (u ⊆ v) -> exists a : A, ~ (a ∈ u -> a ∈ v).
Proof.
  intros ni. destruct (not_incl_nincl ni) as [a iu niv]. exists a. tauto.
Qed.

(* ------------------------------------------------------------ *)
(* Strict inclusion *)
(* Useless non-constructive definition *)
Definition ncsincl {A} (u v : set A) : Prop :=  u ⊆ v /\ ~ v ⊆ u.

(* Simpler than incl_ncsincl! *)
Lemma sincl_sdiff `{ED: EqDec A} {u v w} : w ⊆ u -> v ⊂ w -> (u \ w) ⊂ (u \ v).
Proof.
  intros iwu [ivw [a aw nav]]. split.
  - clear a aw nav. intros a iuw.
    destruct (sdiff_elim iuw) as [iu niw]. apply sdiff_intro; intuition.
  - exists a.
    + apply sdiff_intro; intuition.
    + intro iuw. destruct (sdiff_elim iuw) as [_ niw]. exact (niw aw).
Qed.

Lemma sincl_len_lt `{ED: EqDec A} {u v : set A} :
  NoDup u -> u ⊂ v -> length u < length v.
Proof.
  intros nd [icuv [a iv niu]].
  assert (luv := NoDup_incl_length nd icuv : length u <= length v).
  destruct (Arith.Lt.le_lt_or_eq _ _ luv) as [ltuv | euv]; clear luv;
    try (exact ltuv).
  case niu. assert (icvu : v ⊆ u).
  { apply (NoDup_length_incl nd).
    - rewrite euv. trivial.
    - exact icuv. }
  apply icvu. exact iv.
Qed.

(* ------------------------------------------------------------ *)
(* Filter *)

Lemma filter_intro {A} {f} (a : A) {u} : a ∈ u -> f a = true -> a ∈ (filter f u).
Proof.
  intros i e. apply filter_In. split; assumption.
Qed.

Lemma filter_elim {A} {f} {a : A} {u} : a ∈ (filter f u) -> a ∈ u /\ f a = true.
Proof.
  intro i. apply filter_In. assumption.
Qed.

Lemma filter_empty {A} {a:A} {f u} : a ∈ u -> filter f u = ∅ -> f a = false.
Proof.
  intros al flemp. case_eq (f a).
  - intro fa_true. case (empty_elim a). rewrite <- flemp.
    apply filter_intro; assumption.
  - reflexivity.
Qed.

Lemma filter_incl {A} f (u: list A) : (filter f u) ⊆ u.
Proof.
  intros a i. destruct (filter_elim i) as [iu _]. exact iu.
Qed.

Lemma diff_filter_elim `{ED: EqDec A} {a f u v} :
  a ∈ u \ filter f v -> a ∈ u \ v  \/  f a = false.
Proof.
  intros i. destruct (sdiff_elim i) as [iu nifv].
  case_eq (f a); intro efx.
  - left. apply sdiff_intro.
    + exact iu.
    + intro iv. apply nifv. apply filter_intro; assumption.
  - right; reflexivity.
Qed.

Definition all_false {A} (u : set A) (f : A -> bool)  : Prop :=
  forall a, a ∈ u -> f a = false.

Lemma filter_sdiff_empty `{ED: EqDec A} {f u v} :
  filter f (u \ v) = ∅ -> all_false v f -> all_false u f.
Proof.
  intros fiuv afv a au. red in afv.
  destruct (sdiff_union v au) as [av | auv].
  - apply afv; exact av.
  - apply (filter_empty auv fiuv).
Qed.

(* ------------------------------------------------------------ *)
(* NoDup and nodup *)

(* Remark: the following 
Lemma nodup_incl `{ED: EqDec A} {u v: set A} : u ⊆ v -> NoDup v -> NoDup u.
is WRONG!
Counter example  u = [a; a] and v = [a]
*)

Definition NoDup_empty {A: Type} : NoDup ∅ := List.NoDup_nil A.
Definition NoDup_cons {A: Type} {x l} : x ∉ l -> NoDup l -> NoDup (x :: l) :=
  @List.NoDup_cons A x l.

Lemma NoDup_cons_elim {A} {a :A} {l} : NoDup (a :: l) -> a ∉ l /\ NoDup l.
Proof.
  destruct (NoDup_cons_iff a l) as [ndal _]. exact ndal.
Qed.

Definition nodup `{ED: EqDec A} := List.nodup Aeq_dec.
Definition sadd_nodup `{ED: EqDec A} := ListSet.set_add_nodup Aeq_dec.
Definition union_nodup `{ED: EqDec A} := ListSet.set_union_nodup Aeq_dec.
Definition sdiff_nodup `{ED: EqDec A} := ListSet.set_diff_nodup Aeq_dec.

Lemma nodup_In_intro `{ED: EqDec A} {x u} : x ∈ u -> x ∈ nodup u.
Proof. intro xu. apply nodup_In. exact xu. Qed.

Lemma nodup_In_elim `{ED: EqDec A} {x u} : x ∈ nodup u -> x ∈ u.
Proof. intro xu. apply (nodup_In Aeq_dec u x). exact xu. Qed.

Lemma nodup_inhabited {A} {ED: EqDec A} u : forall (x : A), x ∈ u -> ~ nodup u = ∅.
Proof.
  induction u as [| y u Hu]; intro x.
  - intro F. case F.
  - simpl. intro yxu. destruct (List.in_dec Aeq_dec y u) as [i | i].
    + apply (Hu y). exact i.
    + discriminate.
Qed.

Lemma nodup_empty {A} {ED: EqDec A} (u : set A) : nodup u = ∅ -> u = ∅.
Proof.
  destruct u as [| x u].
  - reflexivity.
  - intro nd_xu. case (nodup_inhabited (x :: u) x).
    + left. reflexivity.
    + exact nd_xu.
Qed.

Lemma inter_nodup_left `{ED: EqDec A} l l' : NoDup l -> NoDup (l ∩ l').
Proof.
  induction l as [| a l IHl]; simpl.
  - constructor.
  - intro ndal. destruct (NoDup_cons_elim ndal) as [ni nd]. specialize (IHl nd).
    case (set_mem Aeq_dec a l'); try exact IHl.
    apply NoDup_cons; try exact IHl.
    intro i; apply ni. destruct (inter_elim i); assumption.
Qed.

Lemma filter_nodup `{ED: EqDec A} f (l : set A) : NoDup l -> NoDup (filter f l).
Proof.
  induction l as [| a l IHl]; simpl.
  - constructor.
  - intro ndal. destruct (NoDup_cons_elim ndal) as [ni nd]. specialize (IHl nd).
    case  (f a); try (exact IHl).
    apply NoDup_cons; try (exact IHl).
    intro i. apply ni. apply filter_In in i. destruct i; assumption.
Qed.

Lemma sdiff_incl_nodup1 `{ED: EqDec A} (u v : set A) : u \ v ⊆ nodup u \ v.
Proof.
  intros x xuv.
  destruct (sdiff_elim xuv) as [xu xv]. apply sdiff_intro.
  - apply nodup_In_intro. assumption.
  - assumption.
Qed.
