(* AUTEURS / AUTHORS :                                     *)
(* Yuxin Deng, Jiatong University, Shanghai                *)
(* Jean-François Monin, Verimag, Université Grenoble Alpes *)
(*                                             2010 - 2019 *)


Require Export QArith_base.

Definition is_0 (x : Q) : bool := Z.eqb 0 (Qnum x).

Definition is_pos (x : Q) : bool := Z.ltb 0 (Qnum x).

Lemma eq_0 : forall a, a == 0 -> is_0 a = true.
Proof.
  intros * e0. unfold is_0. unfold Qeq in e0; simpl in e0.
  rewrite <- (Zred_factor0 (Qnum a)) in e0.
  apply Z.eqb_eq. symmetry. assumption.
Qed.

Lemma pos_0 : forall a, 0 < a -> is_pos a = true.
Proof.
  intros * l0.  unfold is_pos. unfold Qlt in l0; simpl in l0.
  rewrite <- (Zred_factor0 (Qnum a)) in l0.
  apply Z.ltb_lt. assumption.
Qed.

Lemma pos_0_true : forall a, is_pos a = true -> 0 < a.
Proof.
  intros * pa. unfold is_pos in pa.
  unfold Qlt; simpl. apply Z.ltb_lt.
  rewrite <- (Zred_factor0 (Qnum a)). assumption.
Qed.

Lemma pos_0_false : forall a, is_pos a = false -> a <= 0.
Proof.
  intros * pa. destruct (Qlt_le_dec 0 a) as [l0a | la0].
  - rewrite (pos_0 _ l0a) in pa. discriminate pa.
  - exact la0.
Qed.

Lemma pos_neq a b :
  0 <= a -> 0 <= b -> Qeq_bool a b = false -> 0 < a \/ 0 < b.
Proof.
  intros Ha Hb Hi. apply Qle_lteq in Ha; apply Qle_lteq in Hb.
  destruct Ha as [la | ea]; destruct Hb as [lb | eb]; intuition.
  rewrite <-ea in Hi; rewrite <- eb in Hi; discriminate Hi.
Qed.

Lemma neq_sym a b : Qeq_bool a b = false -> Qeq_bool b a = false.
Proof.
  intros e. case_eq (Qeq_bool b a); trivial. intro et.
  case (Qeq_bool_neq _ _ e). apply Qeq_sym. apply (Qeq_bool_eq _ _ et).
Qed.

Lemma Qeq_eq_bool_false x y : ~ x == y -> Qeq_bool x y = false.
Proof.
  intro ne. case_eq (Qeq_bool x y); intro e.
  - case ne. apply Qeq_bool_eq. exact e.
  - reflexivity.
Qed.

(* *)

Lemma mult_l a b c : b == c -> a * b == a * c.
Proof.
  destruct (Qeq_dec a 0) as [e0 | ne0].
  - rewrite e0. intuition.
  - apply (Qmult_inj_l b c _ ne0).
Qed.

Lemma Qeq_bool_false_eq {x1 x2 y1 y2} :
   Qeq_bool x1 x2 = false -> (y1 == y2 -> x1 == x2) -> Qeq_bool y1 y2 = false.
Proof.
  intros xf ee.
  apply not_true_iff_false; intro ey. apply not_true_iff_false in xf. apply xf.
  rewrite Qeq_bool_iff. apply ee. apply Qeq_bool_iff. exact ey.
Qed.

Lemma Qpos_not_null x : 0 < x -> ~ x == 0.
Proof. intros l e. revert l. rewrite e. apply Qlt_irrefl. Qed.

Lemma Qle_neq_lt a b : a <= b -> ~(a == b) -> a < b.
Proof.
  intros lab nab. destruct (Qle_lt_or_eq _ _ lab) as [ltab | eab].
  - assumption.
  - case (nab eab).
Qed.

Lemma mult_strict_pos_intro a b : 0 < a -> 0 < b -> 0 < a * b.
Proof.
  intros la lb. 
  rewrite <- (Qmult_0_l b). apply Qmult_lt_compat_r; assumption.
Qed.

Lemma mult_strict_pos_elim1 a b : 0 <= b -> 0 < a * b -> 0 < a.
Proof.
  intros pb pab. destruct (Q_dec a 0) as [[na | pa] | ea].
  - assert(H1: a*b <= 0*b) by ( apply Qmult_le_compat_r; [intuition | assumption]).
    rewrite Qmult_0_l in H1. 
    case (Qlt_irrefl 0). apply (Qlt_le_trans _ _ _ pab H1).
  - exact pa.
  - case (Qlt_irrefl 0). revert pab. rewrite ea. rewrite Qmult_0_l. trivial.
Qed.

Lemma mult_pos_monot a b c d :
  0 <= a -> 0 <= c -> a <= b -> c <= d -> a * c <= b * d.
Proof.
  intros pa pc lab lcd. apply Qle_trans with (y := b*c).
  - apply Qmult_le_compat_r; assumption.
  - rewrite Qmult_comm. assert (ebc: b * d == d * b) by apply Qmult_comm. rewrite ebc. 
    apply Qmult_le_compat_r.
    + assumption.
    + apply (Qle_trans _ _ _ pa lab).
Qed.

Lemma mult_pos_smonot a b c d :
  0 <= a -> 0 <= c -> a < b -> c < d -> a * c < b * d.
Proof.
  intros pa pc lab lcd. apply Qle_lt_trans with (y := b*c).
  - apply Qmult_le_compat_r.
    + apply Qlt_le_weak. exact lab.
    + exact pc.
  - rewrite Qmult_comm. assert (ebc: b * d == d * b) by apply Qmult_comm. rewrite ebc.
    apply Qmult_lt_compat_r.
    + apply (Qle_lt_trans _ _ _ pa lab).
    + exact lcd.
Qed.

Fixpoint natpow (a : Q) (n : nat) : Q :=
  match n with
  | O => 1
  | S n => a * natpow a n
  end.

Lemma natpow_eq {a b} n : a == b -> natpow a n == natpow b n.
Proof.
  intro e. induction n as [| n IHn]; cbn.
  - reflexivity.
  - rewrite IHn. rewrite e. reflexivity.
Qed.

Lemma Qpow_spos n a : 0 < a -> 0 < natpow a n.
Proof.
  intro pos. induction n as [| n IHn]; cbn.
  - apply pos_0_true. reflexivity.
  - apply mult_strict_pos_intro; trivial.
Qed.

Lemma Qpow_pos n a : 0 <= a -> 0 <= natpow a n.
Proof.
  intro pos. induction n as [| n IHn]; cbn.
  - apply Qle_bool_imp_le. reflexivity.
  - apply Qmult_le_0_compat; trivial.
Qed.

Lemma Qpow_monot n a b :
  0 <= a -> a <= b -> natpow a n <= natpow b n.
Proof.
  intros pa lab. induction n as [| n IHn]; simpl.
  - apply Qle_refl.
  - apply mult_pos_monot; try assumption. apply Qpow_pos. assumption.
Qed.

Lemma Qpow_smonot n a b :
  0 <= a -> a < b -> natpow a (S n) < natpow b (S n).
Proof.
  intros pa lab. cbn. induction n as [| n IHn]; simpl.
  - do 2 rewrite Qmult_1_r. exact lab.
  - apply mult_pos_smonot; try assumption. apply (Qpow_pos (S n)). exact pa.
Qed.

Lemma pow_root n a b :
  0 <= a -> 0 <= b -> 
  natpow a (S n) == natpow b (S n) ->
  a == b.
Proof.
  intros pa pb e.
  destruct (Q_dec a b) as [[lab | lba] | eab]. 
  + assert (e' := Qpow_smonot n _ _ pa lab). rewrite e in e'.
    case (Qlt_irrefl _ e').
  + assert (e' := Qpow_smonot n _ _ pb lba). rewrite e in e'.
    case (Qlt_irrefl _ e').
  + assumption.
Qed.
